""" Exceptions defined by PyNMR-STAR. """


class ParsingError(ValueError):
    """ Indicates that something went wrong when parsing NMR-STAR data.
    A line number on which the exception occurred will be provided if
    possible. """

    def __init__(self, message, line_number: int = None):
        Exception.__init__(self)
        self.message = message
        self.line_number = line_number

    def __repr__(self) -> str:
        if self.line_number is not None:
            return f'ParsingError("{self.message}") on line {self.line_number}'
        else:
            return f'ParsingError("{self.message}")'

    def __str__(self) -> str:
        if self.line_number is not None:
            return f"{self.message} Error detected on line {self.line_number}."
        else:
            return self.message


class InvalidStateError(ValueError):
    """ Indicates that the data as exists in the PyNMRSTAR objects is not
    consistent with the NMR-STAR format or dictionary. This often means that
    internal PyNMRSTAR attributes were modified directly rather than using the
    appropriate getters/setters, since the library attempts to prevent actions
    which would lead to such states. """

    def __init__(self, message):
        Exception.__init__(self)
        self.message = message

    def __repr__(self) -> str:
        return f'InvalidStateError("{self.message}")'

    def __str__(self) -> str:
        return self.message




import logging
import re
from typing import Optional

from pynmrstar import definitions, cnmrstar, entry as entry_mod, loop as loop_mod, saveframe as saveframe_mod
from pynmrstar.exceptions import ParsingError

logger = logging.getLogger('pynmrstar')


class Parser(object):
    """Parses an entry. You should not ever use this class directly."""

    def __init__(self, entry_to_parse_into: 'entry_mod.Entry' = None) -> None:

        # Just make an entry to parse into if called with no entry passed
        if entry_to_parse_into is None:
            entry_to_parse_into = entry_mod.Entry.from_scratch("")

        self.ent: entry_mod.Entry = entry_to_parse_into
        self.full_data: str = ""
        self.token: str = ""
        self.source: str = "unknown"
        self.delimiter: str = " "
        self.line_number: int = 0

    def get_token(self) -> str:
        """ Returns the next token in the parsing process."""

        try:
            self.token, self.line_number, self.delimiter = cnmrstar.get_token_full()
        except ValueError as err:
            raise ParsingError(str(err))

        return self.token

    @staticmethod
    def load_data(data: str) -> None:
        """ Loads data in preparation of parsing and cleans up newlines
        and massages the data to make parsing work properly when multi-line
        values aren't as expected. Useful for manually getting tokens from
        the parser."""

        # Fix DOS line endings
        data = data.replace("\r\n", "\n").replace("\r", "\n")
        # Change '\n; data ' started multi-lines to '\n;\ndata'
        data = re.sub(r'\n;([^\n]+?)\n', r'\n;\n\1\n', data)

        cnmrstar.load_string(data)

    def parse(self, data: str, source: str = "unknown", raise_parse_warnings: bool = False,
              convert_data_types: bool = False) -> 'entry_mod.Entry':
        """ Parses the string provided as data as an NMR-STAR entry
        and returns the parsed entry. Raises ParsingError on exceptions.

        Set raise_parse_warnings to raise an exception if the file has
        something technically incorrect, but still parsable.

        Following is a list of the types of errors that would trigger
        raise_parse_warnings:

        * A loop with no data was found.
        * A loop with no tags or values was found.
        * A tag with an improper multi-line value was found.
        Multi-line values should look like this:
        \n;\nThe multi-line\nvalue here.\n;\n
        but the tag looked like this:
        \n; The multi-line\nvalue here.\n;\n"""

        self.load_data(data)
        self.get_token()

        # Make sure this is actually a STAR file
        if not self.token.lower().startswith("data_"):
            raise ParsingError("Invalid file. NMR-STAR files must start with 'data_' followed by the data name. "
                               f"Did you accidentally select the wrong file? Your file started with '{self.token}'.",
                               self.line_number)

        # Make sure there is a data name
        elif len(self.token) < 6:
            raise ParsingError("'data_' must be followed by data name. Simply 'data_' is not allowed.",
                               self.line_number)

        if self.delimiter != " ":
            raise ParsingError("The data_ keyword may not be quoted or semicolon-delimited.",
                               self.line_number)

        # Set the entry_id
        self.ent._entry_id = self.token[5:]
        self.source = source

        # We are expecting to get saveframes
        while self.get_token() is not None:

            if not self.token.lower().startswith("save_"):
                raise ParsingError(f"Only 'save_NAME' is valid in the body of a NMR-STAR file. Found '{self.token}'.",
                                   self.line_number)

            if len(self.token) < 6:
                raise ParsingError("'save_' must be followed by saveframe name. You have a 'save_' tag which is "
                                   "illegal without a specified saveframe name.", self.line_number)

            if self.delimiter != " ":
                raise ParsingError("The save_ keyword may not be quoted or semicolon-delimited.",
                                   self.line_number)

            # Add the saveframe
            cur_frame: Optional[saveframe_mod.Saveframe] = saveframe_mod.Saveframe.from_scratch(self.token[5:],
                                                                                                source=source)
            self.ent.add_saveframe(cur_frame)

            # We are in a saveframe
            while self.get_token() is not None:

                if self.token.lower() == "loop_":
                    if self.delimiter != " ":
                        raise ParsingError("The loop_ keyword may not be quoted or semicolon-delimited.",
                                           self.line_number)

                    cur_loop: Optional[loop_mod.Loop] = loop_mod.Loop.from_scratch(source=source)

                    # We are in a loop
                    cur_data = []
                    seen_data = False
                    in_loop = True
                    while in_loop and self.get_token() is not None:

                        # Add a tag if it isn't quoted - if quoted, it should be treated as a data value
                        if self.token.startswith("_") and self.delimiter == " ":
                            try:
                                cur_loop.add_tag(self.token)
                            except ValueError as err:
                                raise ParsingError(str(err), self.line_number)

                        # On to data
                        else:

                            # Now that we have the tags we can add the loop
                            #  to the current saveframe
                            try:
                                cur_frame.add_loop(cur_loop)
                            except ValueError as err:
                                raise ParsingError(str(err), self.line_number)

                            # We are in the data block of a loop
                            while self.token is not None:
                                if self.token.lower() == "stop_":
                                    if self.delimiter != " ":
                                        raise ParsingError(
                                            "The stop_ keyword may not be quoted or semicolon-delimited.",
                                            self.line_number)
                                    if len(cur_loop.tags) == 0:
                                        if raise_parse_warnings:
                                            raise ParsingError("Loop with no tags.", self.line_number)
                                        else:
                                            logger.warning('Loop with no tags in parsed file on line: %s',
                                                           self.line_number)
                                        cur_loop = None
                                    if not seen_data:
                                        if raise_parse_warnings:
                                            raise ParsingError("Loop with no data.", self.line_number)
                                        else:
                                            logger.warning("Loop with no data on line: %s", self.line_number)

                                    if len(cur_data) > 0:
                                        if len(cur_data) % len(cur_loop.tags) != 0:
                                            raise ParsingError(f"The loop being parsed, '{cur_loop.category}' does "
                                                               f"not have the expected number of data elements. This "
                                                               f"indicates that either one or more tag values are "
                                                               f"either missing from or duplicated in this loop.",
                                                               self.line_number)
                                        try:
                                            cur_loop.add_data(cur_data, rearrange=True,
                                                              convert_data_types=convert_data_types)
                                        # If there is an issue with the loops during parsing, raise a parse error
                                        #  rather than the ValueError that would be raised if they made the mistake
                                        #   directly
                                        except ValueError as e:
                                            raise ParsingError(str(e))
                                    cur_data = []

                                    cur_loop = None
                                    in_loop = False
                                    break
                                elif self.token.startswith("_") and self.delimiter == " ":
                                    raise ParsingError("Cannot have more loop tags after loop data. Or perhaps this "
                                                       f"was a data value which was not quoted (but must be, "
                                                       f"if it starts with '_')? Value: '{self.token}'.",
                                                       self.line_number)
                                else:
                                    if len(cur_loop.tags) == 0:
                                        raise ParsingError("Data value found in loop before any loop tags were "
                                                           "defined. Value: '{self.token}'",
                                                           self.line_number)

                                    if self.token.lower() in definitions.RESERVED_KEYWORDS and self.delimiter == " ":
                                        error = "Cannot use keywords as data values unless quoted or semi-colon " \
                                                "delimited. Perhaps this is a loop that wasn't properly terminated " \
                                                "with a 'stop_' keyword before the saveframe ended or another loop " \
                                                f"began? Value found where 'stop_' or another data value expected: " \
                                                f"'{self.token}'."
                                        if len(cur_data) > 0:
                                            error += f" Last loop data element parsed: '{cur_data[-1]}'."
                                        raise ParsingError(error, self.line_number)
                                    cur_data.append(self.token)
                                    seen_data = True

                                # Get the next token
                                self.get_token()

                    if not self.token:
                        raise ParsingError(f"Loop improperly terminated at end of file. Loops must end with the "
                                           f"'stop_' token, but the file ended without the stop token.",
                                           self.line_number)
                    if self.token.lower() != 'stop_':
                        raise ParsingError(f"Loop improperly terminated at end of file. Loops must end with the "
                                           f"'stop_' token, but the token '{self.token}' was found instead.",
                                           self.line_number)

                # Close saveframe
                elif self.token.lower() == "save_":
                    if self.delimiter not in " ;":
                        raise ParsingError("The save_ keyword may not be quoted or semicolon-delimited.",
                                           self.line_number)

                    if cur_frame.tag_prefix is None:
                        raise ParsingError("The tag prefix was never set! Either the saveframe had no tags, you "
                                           "tried to read a version 2.1 file, or there is something else wrong with "
                                           f"your file. Saveframe error occurred within: '{cur_frame.name}'",
                                           line_number=self.line_number)
                    break

                # Invalid content in saveframe
                elif not self.token.startswith("_"):
                    if cur_frame.name == 'internaluseyoushouldntseethis_frame':
                        raise ParsingError(f"Invalid token found in loop contents. Expecting 'loop_' "
                                           f"but found: '{self.token}'", line_number=self.line_number)
                    else:
                        raise ParsingError(f"Invalid token found in saveframe '{cur_frame.name}'. Expecting a tag, "
                                           f"loop, or 'save_' token but found: '{self.token}'",
                                           line_number=self.line_number)

                # Add a tag
                else:
                    if self.delimiter != " ":
                        raise ParsingError(f"Saveframe tags may not be quoted or semicolon-delimited. Quoted tag: '"
                                           f"{self.token}'.",
                                           self.line_number)
                    cur_tag: Optional[str] = self.token

                    # We are in a saveframe and waiting for the saveframe tag
                    self.get_token()
                    if self.delimiter == " ":
                        if self.token.lower() in definitions.RESERVED_KEYWORDS:
                            raise ParsingError("Cannot use keywords as data values unless quoted or semi-colon "
                                               f"delimited. Illegal value: '{self.token}'", self.line_number)
                        if self.token.startswith("_"):
                            raise ParsingError(
                                "Cannot have a tag value start with an underscore unless the entire value "
                                "is quoted. You may be missing a data value on the previous line. "
                                f"Illegal value: '{self.token}'", self.line_number)
                    try:
                        cur_frame.add_tag(cur_tag, self.token, convert_data_types=convert_data_types)
                    except ValueError as err:
                        raise ParsingError(str(err), line_number=self.line_number)

            if not self.token or self.token.lower() != "save_":
                raise ParsingError("Saveframe improperly terminated at end of file. Saveframes must be terminated "
                                   "with the 'save_' token.",
                                   self.line_number)

        # Free the memory of the original copy of the data we parsed
        self.full_data = None

        # Reset the parser
        cnmrstar.reset()

        return self.ent



#!/usr/bin/env python3

"""This module provides :py:class:`pynmrstar.Entry`, :py:class:`pynmrstar.Saveframe`,
   :py:class:`pynmrstar.Loop`, and :py:class:`pynmrstar.Schema` objects.

It also provides some utility functions in :py:obj:`pynmrstar.utils`

Use python's built in help function for documentation."""

import decimal as _decimal
import logging
import os

try:
    import cnmrstar
except ImportError:
    try:
        import pynmrstar.cnmrstar as cnmrstar
    except ImportError:
        if os.environ.get('READTHEDOCS'):
            cnmrstar = None
        else:
            raise ImportError('Could not import cnmrstar sub-module! Your installation appears to be broken.')

from pynmrstar import utils
from pynmrstar._internal import __version__, min_cnmrstar_version
from pynmrstar.entry import Entry
from pynmrstar.loop import Loop
from pynmrstar.parser import Parser as _Parser
from pynmrstar.saveframe import Saveframe
from pynmrstar.schema import Schema
import pynmrstar.definitions as definitions

if cnmrstar:
    if "version" not in dir(cnmrstar):
        raise ImportError(f"Could not determine the version of cnmrstar installed, and version {min_cnmrstar_version} or "
                          "greater is required.")
    if cnmrstar.version() < min_cnmrstar_version:
        raise ImportError("The version of the cnmrstar module installed does not meet the requirements. As this should be "
                          f"handled automatically, there may be an issue with your installation. Version installed: "
                          f"{cnmrstar.version()}. Version required: {min_cnmrstar_version}")

# Set up logging
logger = logging.getLogger('pynmrstar')

# This makes sure that when decimals are printed a lower case "e" is used
_decimal.getcontext().capitals = 0

del loop
del entry
del saveframe
del schema
del parser

__all__ = ['Loop', 'Saveframe', 'Entry', 'Schema', 'definitions', 'utils', '__version__', 'exceptions', 'cnmrstar']



import decimal
import json
import logging
import os
import time
import zlib
from datetime import date
from gzip import GzipFile
from io import StringIO, BytesIO
from typing import Dict, Union, IO, List, Tuple
from urllib.error import HTTPError, URLError
from urllib.request import urlopen, Request

import pynmrstar

__version__: str = "3.3.1"
min_cnmrstar_version: str = "3.2.0"

# If we have requests, open a session to reuse for the duration of the program run
try:
    from requests import session as _requests_session
    # This replaces the urllib HTTPError if we have requests
    from requests.exceptions import HTTPError, ConnectionError
    _session = _requests_session()
except ModuleNotFoundError:
    _session = None

logger = logging.getLogger('pynmrstar')


# noinspection PyDefaultArgument
def _get_comments(_comment_cache: Dict[str, Dict[str, str]] = {}) -> Dict[str, Dict[str, str]]:
    """ Loads the comments that should be placed in written files.

    The default argument is mutable on purpose, as it is used as a cache for memoization."""

    # Comment dictionary already exists
    if _comment_cache:
        return _comment_cache

    file_to_load = os.path.join(os.path.dirname(os.path.realpath(__file__)))
    file_to_load = os.path.join(file_to_load, "reference_files/comments.str")

    # The import needs to be here to avoid import errors due to circular imports
    from pynmrstar.entry import Entry
    try:
        comment_entry = Entry.from_file(file_to_load)
    except IOError:
        # Load the comments from Github if we can't find them locally
        try:
            logger.warning('Could not load comments from disk. Loading from web...')
            comment_entry = Entry.from_file(_interpret_file(pynmrstar.definitions.COMMENT_URL))
        except Exception:
            logger.exception('Could not load comments from web. No comments will be shown.')
            # No comments will be printed
            return {}

    # Load the comments
    comment_records = comment_entry[0][0].get_tag(["category", "comment", "every_flag"])
    comment_map = {'N': False, 'Y': True}
    for comment in comment_records:
        if comment[1] != ".":
            _comment_cache[comment[0]] = {'comment': comment[1].rstrip() + "\n\n",
                                          'every_flag': comment_map[comment[2]]}

    return _comment_cache


def _json_serialize(obj: object) -> str:
    """JSON serializer for objects not serializable by default json code"""

    # Serialize datetime.date objects by calling str() on them
    if isinstance(obj, (date, decimal.Decimal)):
        return str(obj)
    raise TypeError("Type not serializable: %s" % type(obj))


def _get_url_reliably(url: str, wait_time: float = 10, raw: bool = False, timeout: int = 10, retries: int = 2):
    """ Attempts to load data from a URL, retrying the specified number of times with an exponential
    backoff if rate limited. Fails immediately on 4xx errors that are not 403."""

    global _session

    # If using Requests
    if _session:
        try:
            response = _session.get(url, timeout=timeout,
                                    headers={'Application': f'PyNMRSTAR {__version__}'})
        except ConnectionError:
            _session = _requests_session()
            try:
                response = _session.get(url, timeout=timeout,
                                        headers={'Application': f'PyNMRSTAR {__version__}'})
            except ConnectionError:
                raise HTTPError("A ConnectionError was thrown during an attempt to load the entry.")

        # We are rate limited - sleep and try again
        if response.status_code == 403:
            if retries > 0:
                logger.warning(f'We were rate limited. Sleeping for {wait_time} seconds.')
                time.sleep(wait_time)
                return _get_url_reliably(url, wait_time=wait_time*2, raw=raw, timeout=timeout,
                                         retries=retries - 1)
            else:
                raise HTTPError("Continued to receive 403 (forbidden, due to rate limit) after multiple wait times.") \
                    from None
        if response.status_code == 404:
            raise KeyError(f"Server returned 404.") from None
        response.raise_for_status()
        if raw:
            return response.content
        else:
            return response.text
    else:
        # Use the built in library
        try:
            req = Request(url)
            req.add_header('Application', f'PyNMRSTAR {__version__}')
            url_request = urlopen(req, timeout=timeout)
            serialized_ent = url_request.read()
            url_request.close()

        except HTTPError as err:
            if err.code == 404:
                raise KeyError(f"Server returned 404.") from None
            # We are rate limited - sleep and try again
            elif err.code == 403:
                if retries > 0:
                    logger.warning(f'We were rate limited. Sleeping for {wait_time} seconds.')
                    time.sleep(wait_time)
                    return _get_url_reliably(url, wait_time=wait_time * 2, raw=raw, timeout=timeout,
                                             retries=retries - 1)
                else:
                    raise HTTPError("Continued to receive 403 (forbidden, due to rate limit) after multiple wait "
                                    "times.") from None
            else:
                raise err
        if raw:
            return serialized_ent
        else:
            return serialized_ent.decode()


def _get_entry_from_database(entry_num: Union[str, int], convert_data_types: bool = False) -> 'pynmrstar.Entry':
    """ Fetches an entry from the API (or falls back to the FTP site) in
    as reliable and robust a way as possible. Used by Entry.from_database(). """

    entry_num = str(entry_num).lower()
    if entry_num.startswith("bmr"):
        entry_num = entry_num[3:]

    # Try to load the entry using JSON

    entry_url: str = (pynmrstar.definitions.API_URL + "/entry/%s?format=zlib") % entry_num

    try:
        serialized_ent = _get_url_reliably(entry_url, raw=True, retries=2)
        json_data = json.loads(zlib.decompress(serialized_ent).decode())
        if "error" in json_data:
            raise RuntimeError('Something wrong with API response.')
        ent = pynmrstar.Entry.from_json(json_data)
    except (HTTPError, ConnectionError, RuntimeError):
        # Can't fall back to FTP for chemcomps
        if entry_num.startswith("chemcomp"):
            raise IOError("Unable to load that chemcomp from the API.")

        # We're going to try again from the FTP
        logger.warning('Failed to download entry from the API, trying again from the FTP site.')
        if "bmse" in entry_num or "bmst" in entry_num:
            url = f"{pynmrstar.definitions.FTP_URL}/metabolomics/entry_directories/{entry_num}/{entry_num}.str"
        else:
            url = f"{pynmrstar.definitions.FTP_URL}/entry_directories/bmr{entry_num}/bmr{entry_num}_3.str"
        try:
            # Use a longer timeout for the timeout
            entry_content = _get_url_reliably(url, raw=False, timeout=20, retries=1)
            ent = pynmrstar.Entry.from_string(entry_content)
        except HTTPError:
            raise IOError(f"Entry {entry_num} does not exist in the public database.") from None
        except URLError:
            raise IOError("You don't appear to have an active internet connection. Cannot fetch entry.") from None

    except KeyError:
        raise IOError(f"Entry {entry_num} does not exist in the public database.") from None

    # Update the entry source
    ent.source = f"from_database({entry_num})"
    for each_saveframe in ent:
        each_saveframe.source = ent.source
        for each_loop in each_saveframe:
            each_loop.source = ent.source

    if convert_data_types:
        schema = pynmrstar.utils.get_schema()
        for each_saveframe in ent:
            for tag in each_saveframe.tags:
                cur_tag = each_saveframe.tag_prefix + "." + tag[0]
                tag[1] = schema.convert_tag(cur_tag, tag[1])
            for loop in each_saveframe:
                for row in loop.data:
                    for pos in range(0, len(row)):
                        category = loop.category + "." + loop.tags[pos]
                        row[pos] = schema.convert_tag(category, row[pos])

    return ent


def _interpret_file(the_file: Union[str, IO]) -> StringIO:
    """Helper method returns some sort of object with a read() method.
    the_file could be a URL, a file location, a file object, or a
    gzipped version of any of the above."""

    if hasattr(the_file, 'read'):
        read_data: Union[bytes, str] = the_file.read()
        if type(read_data) == bytes:
            buffer: BytesIO = BytesIO(read_data)
        elif type(read_data) == str:
            buffer = BytesIO(read_data.encode())
        else:
            raise IOError("What did your file object return when .read() was called on it?")
    elif isinstance(the_file, str):
        if the_file.startswith("http://") or the_file.startswith("https://") or the_file.startswith("ftp://"):
            buffer = BytesIO(_get_url_reliably(the_file, raw=True, retries=0))
        else:
            with open(the_file, 'rb') as read_file:
                buffer = BytesIO(read_file.read())
    else:
        raise ValueError("Cannot figure out how to interpret the file you passed.")

    # Decompress the buffer if we are looking at a gzipped file
    try:
        gzip_buffer = GzipFile(fileobj=buffer)
        gzip_buffer.readline()
        gzip_buffer.seek(0)
        buffer = BytesIO(gzip_buffer.read())
    # Apparently we are not looking at a gzipped file
    except (IOError, AttributeError, UnicodeDecodeError):
        pass

    buffer.seek(0)
    return StringIO(buffer.read().decode().replace("\r\n", "\n").replace("\r", "\n"))


def get_clean_tag_list(item: Union[str, List[str], Tuple[str]]) -> List[Dict[str, str]]:
    """ Converts the provided item to a list of dictionaries of
    {
     formatted -> just the lower case tag name (category stripped)
     original -> whatever was provided, completely unmodified
    }"""

    if not isinstance(item, (str, list, tuple)):
        raise ValueError('Invalid object provided. Only a tag name (str), or list of tags (list or tuple)'
                         ' are valid inputs to this function.')

    if isinstance(item, list):
        tag_list: List[str] = item
    elif isinstance(item, tuple):
        tag_list = list(item)
    elif isinstance(item, str):
        tag_list = [item]
    else:
        raise ValueError(f'The value you provided was not a string, list, or tuple. Item: {repr(item)}')

    try:
        return [{"formatted": pynmrstar.utils.format_tag_lc(_), "original": _} for _ in tag_list]
    except AttributeError:
        raise ValueError('Your list or tuple may only contain tag names expressed as strings.')


def write_to_file(nmrstar_object: Union['pynmrstar.Entry', 'pynmrstar.Saveframe'],
                  file_name: str,
                  format_: str = "nmrstar",
                  show_comments: bool = True,
                  skip_empty_loops: bool = False,
                  skip_empty_tags: bool = False):
    """ Writes the object to the specified file in NMR-STAR format. """

    if format_ not in ["nmrstar", "json"]:
        raise ValueError("Invalid output format.")

    data_to_write = ''
    if format_ == "nmrstar":
        data_to_write = nmrstar_object.format(show_comments=show_comments,
                                              skip_empty_loops=skip_empty_loops,
                                              skip_empty_tags=skip_empty_tags)
    elif format_ == "json":
        data_to_write = nmrstar_object.get_json()

    out_file = open(file_name, "w")
    out_file.write(data_to_write)
    out_file.close()



#!/usr/bin/env python3

import pynmrstar
import logging


def remove_software_specific_data(nef_file):
    nef_tags = ['nef_nmr_meta_data', 'nef_molecular_system', 'nef_chemical_shift_list', 'nef_distance_restraint_list',
                'nef_dihedral_restraint_list', 'nef_rdc_restraint_list', 'nef_nmr_spectrum', 'nef_peak_restraint_links',
                'nef_related_entries', 'nef_program_script', 'nef_run_history', 'nef_sequence', 'nef_covalent_links',
                'nef_chemical_shift', 'nef_distance_restraint', 'nef_dihedral_restraint', 'nef_rdc_restraint',
                'nef_spectrum_dimension', 'nef_spectrum_dimension_transfer', 'nef_peak', 'nef_peak_restraint_link',
                'audit', '_nef_program_script.program_name', '_nef_program_script.script_name',
                '_nef_program_script.script',
                '_nef_run_history.run_number', '_nef_run_history.program_name', '_nef_run_history.program_version',
                '_nef_run_history.script_name', '_nef_run_history.script', '_nef_nmr_meta_data.sf_category',
                '_nef_nmr_meta_data.sf_framecode', '_nef_nmr_meta_data.format_name',
                '_nef_nmr_meta_data.format_version',
                '_nef_nmr_meta_data.program_name', '_nef_nmr_meta_data.program_version',
                '_nef_nmr_meta_data.creation_date',
                '_nef_nmr_meta_data.uuid', '_nef_nmr_meta_data.coordinate_file_name',
                '_nef_related_entries.database_name',
                '_nef_related_entries.database_accession_code', '_nef_molecular_system.sf_category',
                '_nef_molecular_system.sf_framecode', '_nef_sequence.index', '_nef_sequence.chain_code',
                '_nef_sequence.sequence_code', '_nef_sequence.residue_name', '_nef_sequence.linking',
                '_nef_sequence.residue_variant', '_nef_sequence.cis_peptide', '_nef_covalent_links.chain_code_1',
                '_nef_covalent_links.sequence_code_1', '_nef_covalent_links.residue_name_1',
                '_nef_covalent_links.atom_name_1', '_nef_covalent_links.chain_code_2',
                '_nef_covalent_links.sequence_code_2',
                '_nef_covalent_links.residue_name_2', '_nef_covalent_links.atom_name_2',
                '_nef_chemical_shift_list.sf_category', '_nef_chemical_shift_list.sf_framecode',
                '_nef_chemical_shift_list.atom_chemical_shift_units', '_nef_chemical_shift.chain_code',
                '_nef_chemical_shift.sequence_code', '_nef_chemical_shift.residue_name',
                '_nef_chemical_shift.atom_name', '_nef_chemical_shift.element', '_nef_chemical_shift.isotope_number',
                '_nef_chemical_shift.value', '_nef_chemical_shift.value_uncertainty',
                '_nef_distance_restraint_list.sf_category', '_nef_distance_restraint_list.sf_framecode',
                '_nef_distance_restraint_list.potential_type', '_nef_distance_restraint_list.restraint_origin',
                '_nef_distance_restraint.index', '_nef_distance_restraint.restraint_id',
                '_nef_distance_restraint.restraint_combination_id', '_nef_distance_restraint.chain_code_1',
                '_nef_distance_restraint.sequence_code_1', '_nef_distance_restraint.residue_name_1',
                '_nef_distance_restraint.atom_name_1', '_nef_distance_restraint.chain_code_2',
                '_nef_distance_restraint.sequence_code_2', '_nef_distance_restraint.residue_name_2',
                '_nef_distance_restraint.atom_name_2', '_nef_distance_restraint.weight',
                '_nef_distance_restraint.target_value', '_nef_distance_restraint.target_value',
                '_nef_distance_restraint.target_value_uncertainty', '_nef_distance_restraint.lower_linear_limit',
                '_nef_distance_restraint.lower_limit', '_nef_distance_restraint.upper_limit',
                '_nef_distance_restraint.upper_linear_limit', '_nef_dihedral_restraint_list.sf_category',
                '_nef_dihedral_restraint_list.sf_framecode', '_nef_dihedral_restraint_list.potential_type',
                '_nef_dihedral_restraint_list.restraint_origin', '_nef_dihedral_restraint.index',
                '_nef_dihedral_restraint.restraint_id', '_nef_dihedral_restraint.restraint_combination_id',
                '_nef_dihedral_restraint.chain_code_1', '_nef_dihedral_restraint.sequence_code_1',
                '_nef_dihedral_restraint.residue_name_1', '_nef_dihedral_restraint.atom_name_1',
                '_nef_dihedral_restraint.chain_code_2', '_nef_dihedral_restraint.sequence_code_2',
                '_nef_dihedral_restraint.residue_name_2', '_nef_dihedral_restraint.atom_name_2',
                '_nef_dihedral_restraint.chain_code_3', '_nef_dihedral_restraint.sequence_code_3',
                '_nef_dihedral_restraint.residue_name_3', '_nef_dihedral_restraint.atom_name_3',
                '_nef_dihedral_restraint.chain_code_4', '_nef_dihedral_restraint.sequence_code_4',
                '_nef_dihedral_restraint.residue_name_4', '_nef_dihedral_restraint.atom_name_4',
                '_nef_dihedral_restraint.weight', '_nef_dihedral_restraint.target_value',
                '_nef_dihedral_restraint.target_value_uncertainty', '_nef_dihedral_restraint.lower_linear_limit',
                '_nef_dihedral_restraint.lower_limit', '_nef_dihedral_restraint.upper_limit',
                '_nef_dihedral_restraint.upper_linear_limit', '_nef_dihedral_restraint.name',
                '_nef_rdc_restraint_list.sf_category', '_nef_rdc_restraint_list.sf_framecode',
                '_nef_rdc_restraint_list.potential_type', '_nef_rdc_restraint_list.restraint_origin',
                '_nef_rdc_restraint_list.tensor_magnitude', '_nef_rdc_restraint_list.tensor_rhombicity',
                '_nef_rdc_restraint_list.tensor_chain_code', '_nef_rdc_restraint_list.tensor_sequence_code',
                '_nef_rdc_restraint_list.tensor_residue_name', '_nef_rdc_restraint.index',
                '_nef_rdc_restraint.restraint_id', '_nef_rdc_restraint.restraint_combination_id',
                '_nef_rdc_restraint.chain_code_1', '_nef_rdc_restraint.sequence_code_1',
                '_nef_rdc_restraint.residue_name_1', '_nef_rdc_restraint.atom_name_1',
                '_nef_rdc_restraint.chain_code_2', '_nef_rdc_restraint.sequence_code_2',
                '_nef_rdc_restraint.residue_name_2', '_nef_rdc_restraint.atom_name_2',
                '_nef_rdc_restraint.weight', '_nef_rdc_restraint.target_value',
                '_nef_rdc_restraint.target_value', '_nef_rdc_restraint.target_value_uncertainty',
                '_nef_rdc_restraint.target_value_uncertainty', '_nef_rdc_restraint.lower_linear_limit',
                '_nef_rdc_restraint.lower_limit', '_nef_rdc_restraint.upper_limit',
                '_nef_rdc_restraint.upper_linear_limit', '_nef_rdc_restraint.scale',
                '_nef_rdc_restraint.distance_dependent', '_nef_nmr_spectrum.sf_category',
                '_nef_nmr_spectrum.sf_framecode', '_nef_nmr_spectrum.num_dimensions',
                '_nef_nmr_spectrum.chemical_shift_list', '_nef_nmr_spectrum.experiment_classification',
                '_nef_nmr_spectrum.experiment_type', '_nef_spectrum_dimension.dimension_id',
                '_nef_spectrum_dimension.axis_unit', '_nef_spectrum_dimension.axis_code',
                '_nef_spectrum_dimension.spectrometer_frequency', '_nef_spectrum_dimension.spectral_width',
                '_nef_spectrum_dimension.value_first_point', '_nef_spectrum_dimension.folding',
                '_nef_spectrum_dimension.absolute_peak_positions', '_nef_spectrum_dimension.is_acquisition',
                '_nef_spectrum_dimension_transfer.dimension_1', '_nef_spectrum_dimension_transfer.dimension_2',
                '_nef_spectrum_dimension_transfer.transfer_type', '_nef_spectrum_dimension_transfer.is_indirect',
                '_nef_peak.index', '_nef_peak.peak_id', '_nef_peak.volume',
                '_nef_peak.volume_uncertainty', '_nef_peak.height', '_nef_peak.height_uncertainty',
                '_nef_peak.position_1', '_nef_peak.position_uncertainty_1', '_nef_peak.position_2',
                '_nef_peak.position_uncertainty_2', '_nef_peak.position_3', '_nef_peak.position_uncertainty_3',
                '_nef_peak.position_4', '_nef_peak.position_uncertainty_4', '_nef_peak.position_5',
                '_nef_peak.position_uncertainty_5', '_nef_peak.position_6', '_nef_peak.position_uncertainty_6',
                '_nef_peak.position_7', '_nef_peak.position_uncertainty_7', '_nef_peak.position_8',
                '_nef_peak.position_uncertainty_8', '_nef_peak.position_9', '_nef_peak.position_uncertainty_9',
                '_nef_peak.position_10', '_nef_peak.position_uncertainty_10', '_nef_peak.position_11',
                '_nef_peak.position_uncertainty_11', '_nef_peak.position_12', '_nef_peak.position_uncertainty_12',
                '_nef_peak.position_13', '_nef_peak.position_uncertainty_13', '_nef_peak.position_14',
                '_nef_peak.position_uncertainty_14', '_nef_peak.position_15', '_nef_peak.position_uncertainty_15',
                '_nef_peak.chain_code_1', '_nef_peak.sequence_code_1', '_nef_peak.residue_name_1',
                '_nef_peak.atom_name_1', '_nef_peak.chain_code_2', '_nef_peak.sequence_code_2',
                '_nef_peak.residue_name_2',
                '_nef_peak.atom_name_2', '_nef_peak.chain_code_3', '_nef_peak.sequence_code_3',
                '_nef_peak.residue_name_3',
                '_nef_peak.atom_name_3', '_nef_peak.chain_code_4', '_nef_peak.sequence_code_4',
                '_nef_peak.residue_name_4',
                '_nef_peak.atom_name_4', '_nef_peak.chain_code_5', '_nef_peak.sequence_code_5',
                '_nef_peak.residue_name_5',
                '_nef_peak.atom_name_5', '_nef_peak.chain_code_6', '_nef_peak.sequence_code_6',
                '_nef_peak.residue_name_6',
                '_nef_peak.atom_name_6', '_nef_peak.chain_code_7', '_nef_peak.sequence_code_7',
                '_nef_peak.residue_name_7',
                '_nef_peak.atom_name_7', '_nef_peak.chain_code_8', '_nef_peak.sequence_code_8',
                '_nef_peak.residue_name_8',
                '_nef_peak.atom_name_8', '_nef_peak.chain_code_9', '_nef_peak.sequence_code_9',
                '_nef_peak.residue_name_9',
                '_nef_peak.atom_name_9', '_nef_peak.chain_code_10', '_nef_peak.sequence_code_10',
                '_nef_peak.residue_name_10',
                '_nef_peak.atom_name_10', '_nef_peak.chain_code_11', '_nef_peak.sequence_code_11',
                '_nef_peak.residue_name_11',
                '_nef_peak.atom_name_11', '_nef_peak.chain_code_12', '_nef_peak.sequence_code_12',
                '_nef_peak.residue_name_12',
                '_nef_peak.atom_name_12', '_nef_peak.chain_code_13', '_nef_peak.sequence_code_13',
                '_nef_peak.residue_name_13',
                '_nef_peak.atom_name_13', '_nef_peak.chain_code_14', '_nef_peak.sequence_code_14',
                '_nef_peak.residue_name_14',
                '_nef_peak.atom_name_14', '_nef_peak.chain_code_15', '_nef_peak.sequence_code_15',
                '_nef_peak.residue_name_15',
                '_nef_peak.atom_name_15', '_nef_peak_restraint_links.sf_category',
                '_nef_peak_restraint_links.sf_framecode',
                '_nef_peak_restraint_link.nmr_spectrum_id', '_nef_peak_restraint_link.peak_id',
                '_nef_peak_restraint_link.restraint_list_id', '_nef_peak_restraint_link.restraint_id',
                '_audit.revision_id', '_audit.creation_date', '_audit.creation_method', '_audit.update_record']

    # Suppress the parse notices, but then log everything after
    logger = logging.getLogger('remove_non_nef_data')
    logger.setLevel(logging.ERROR)
    nef_data = pynmrstar.Entry.from_file(nef_file)
    logger.setLevel(logging.DEBUG)

    saveframe_index = 0
    while saveframe_index < len(nef_data):
        saveframe = nef_data[saveframe_index]
        if not saveframe.category.lower().startswith('nef_'):
            nef_data.remove_saveframe(saveframe)
            logger.debug('Deleted entire saveframe: %s', saveframe.category)
            continue

        tags_to_remove = []
        for tag in saveframe.tags:
            if f'_{saveframe.category}.{tag[0]}'.lower() not in nef_tags:
                tags_to_remove.append(tag)
        saveframe.remove_tag(tags_to_remove)
        logger.debug('Deleted saveframe tags: %s', tags_to_remove)

        loop_index = 0
        while loop_index < len(saveframe):
            loop = saveframe[loop_index]
            if not loop.category.lower().startswith('_nef'):
                del (saveframe[loop_index])
                logger.debug('Deleted entire loop: %s', loop.category)
                continue
            else:
                for tag in loop.get_tag_names():
                    if tag.lower() not in nef_tags:
                        logger.debug('Deleted loop tag: %s', tag)
                        loop.remove_tag(tag)
            loop_index += 1
        saveframe_index += 1

    logger.debug('Finished removing saveframes/loops/tags.')
    nef_data.remove_empty_saveframes()
    out_file = 'cleaned.nef'
    with open(out_file, 'w') as wstarfile:
        wstarfile.write(str(nef_data))
    return True


if __name__ == "__main__":
    remove_software_specific_data('gp78CnewHOME15_test.nef')



#!/usr/bin/env python3

""" This file contains various helper functions."""
import functools
import json
import os
from typing import Iterable, Any, Dict
from urllib.error import HTTPError, URLError

from pynmrstar import definitions, cnmrstar, entry as entry_mod
from pynmrstar._internal import _interpret_file
from pynmrstar.schema import Schema

# Set this to allow import * from pynmrstar to work sensibly
__all__ = ['diff', 'format_category', 'format_tag', 'get_schema', 'iter_entries', 'quote_value', 'validate']


def diff(entry1: 'entry_mod.Entry', entry2: 'entry_mod.Entry') -> None:
    """Prints the differences between two entries. Non-equal entries
    will always be detected, but specific differences detected depends
    on the order of entries."""

    diffs = entry1.compare(entry2)
    if len(diffs) == 0:
        print("Identical entries.")
    for difference in diffs:
        print(difference)


@functools.lru_cache(maxsize=1024)
def format_category(tag: str) -> str:
    """Adds a '_' to the front of a tag (if not present) and strips out
    anything after a '.'"""

    if tag:
        if not tag.startswith("_"):
            tag = "_" + tag
        if "." in tag:
            tag = tag[:tag.index(".")]
    return tag


@functools.lru_cache(maxsize=1024)
def format_tag(tag: str) -> str:
    """Strips anything before the '.'"""

    if '.' in tag:
        return tag[tag.index('.') + 1:]
    return tag


@functools.lru_cache(maxsize=1024)
def format_tag_lc(tag: str) -> str:
    """Strips anything before the '.' and makes the tag lowercase. """

    return format_tag(tag.lower())


# noinspection PyDefaultArgument
def get_schema(passed_schema: 'Schema' = None, _cached_schema: Dict[str, Schema] = {}) -> 'Schema':
    """If passed a schema (not None) it returns it. If passed none,
    it checks if the default schema has been initialized. If not
    initialized, it initializes it. Then it returns the default schema."""

    if passed_schema:
        return passed_schema

    if not _cached_schema:

        # Try to load the local file first
        try:
            schema_file = os.path.join(os.path.dirname(os.path.realpath(__file__)))
            schema_file = os.path.join(schema_file, "reference_files/schema.csv")
            _cached_schema['schema'] = Schema(schema_file=schema_file)
        except IOError:
            # Try to load from the internet
            try:
                _cached_schema['schema'] = Schema()
            except (HTTPError, URLError):
                raise ValueError("Could not load a BMRB schema from the internet or from the local repository.")

    return _cached_schema['schema']


def iter_entries(metabolomics: bool = False) -> Iterable['entry_mod.Entry']:
    """ Returns a generator that will yield an Entry object for every
        macromolecule entry in the current BMRB database. Perfect for performing
        an operation across the entire BMRB database. Set `metabolomics=True`
        in order to get all the entries in the metabolomics database."""

    api_url = f"{definitions.API_URL}/list_entries?database=macromolecules"
    if metabolomics:
        api_url = f"{definitions.API_URL}/list_entries?database=metabolomics"

    for entry in json.loads(_interpret_file(api_url).read()):
        yield entry_mod.Entry.from_database(entry)


@functools.lru_cache(maxsize=65536, typed=True)
def quote_value(value: Any) -> str:
    """Automatically quotes the value in the appropriate way. Don't
    quote values you send to this method or they will show up in
    another set of quotes as part of the actual data. E.g.:

    quote_value('"e. coli"') returns '\'"e. coli"\''

    while

    quote_value("e. coli") returns "'e. coli'"

    This will automatically be called on all values when you use a str()
    method (so don't call it before inserting values into tags or loops).

    Be mindful of the value of STR_CONVERSION_DICT as it will effect the
    way the value is converted to a string.

    """

    # Allow manual specification of conversions for booleans, Nones, etc.
    if value in definitions.STR_CONVERSION_DICT:
        if any(isinstance(value, type(x)) for x in definitions.STR_CONVERSION_DICT):
            value = definitions.STR_CONVERSION_DICT[value]

    return cnmrstar.quote_value(value)


def validate(entry_to_validate: 'entry_mod.Entry', schema: 'Schema' = None) -> None:
    """Prints a validation report of an object."""

    validation = entry_to_validate.validate(schema=schema)
    if len(validation) == 0:
        print("No problems found during validation.")
    for pos, err in enumerate(validation):
        print(f"{pos + 1}: {err}")



import json
import warnings
from copy import deepcopy
from csv import reader as csv_reader, writer as csv_writer
from io import StringIO
from itertools import chain
from typing import TextIO, BinaryIO, Union, List, Optional, Any, Dict, Callable, Tuple

from pynmrstar import definitions, utils, entry as entry_mod
from pynmrstar._internal import _json_serialize, _interpret_file
from pynmrstar.exceptions import InvalidStateError
from pynmrstar.parser import Parser
from pynmrstar.schema import Schema


class Loop(object):
    """A BMRB loop object. Create using the class methods, see below."""

    def __contains__(self, item: Any) -> bool:
        """ Check if the loop contains one or more tags. """

        # Prepare for processing
        if isinstance(item, (list, tuple)):
            to_process: List[str] = list(item)
        elif isinstance(item, str):
            to_process = [item]
        else:
            return False

        lc_tags = self._lc_tags
        for tag in to_process:
            if utils.format_tag_lc(tag) not in lc_tags:
                return False
        return True

    def __eq__(self, other) -> bool:
        """Returns True if this loop is equal to another loop, False if
        it is different."""

        if not isinstance(other, Loop):
            return False

        return (self.category, self._tags, self.data) == \
               (other.category, other._tags, other.data)

    def __getitem__(self, item: Union[int, str, List[str], Tuple[str]]) -> list:
        """Get the indicated row from the data array."""

        try:
            return self.data[item]
        except TypeError:
            if isinstance(item, tuple):
                item = list(item)
            return self.get_tag(tags=item)

    def __init__(self, **kwargs) -> None:
        """ You should not directly instantiate a Loop using this method.
            Instead use the class methods:

            :py:meth:`Loop.from_scratch`, :py:meth:`Loop.from_string`,
            :py:meth:`Loop.from_template`, :py:meth:`Loop.from_file`,
            :py:meth:`Loop.from_json`"""

        # Initialize our local variables
        self._tags: List[str] = []
        self.data: List[List[Any]] = []
        self.category: Optional[str] = None
        self.source: str = "unknown"

        star_buffer: StringIO = StringIO("")

        # Update our source if it provided
        if 'source' in kwargs:
            self.source = kwargs['source']

        # Update our category if provided
        if 'category' in kwargs:
            self.category = utils.format_category(kwargs['category'])
            return

        # They initialized us wrong
        if len(kwargs) == 0:
            raise ValueError("You should not directly instantiate a Loop using this method. Instead use the "
                             "class methods: Loop.from_scratch(), Loop.from_string(), Loop.from_template(), "
                             "Loop.from_file(), and Loop.from_json().")

        # Parsing from a string
        if 'the_string' in kwargs:
            # Parse from a string by wrapping it in StringIO
            star_buffer = StringIO(kwargs['the_string'])
            self.source = "from_string()"
        # Parsing from a file
        elif 'file_name' in kwargs:
            star_buffer = _interpret_file(kwargs['file_name'])
            self.source = f"from_file('{kwargs['file_name']}')"
        # Creating from template (schema)
        elif 'tag_prefix' in kwargs:

            tags = Loop._get_tags_from_schema(kwargs['tag_prefix'], all_tags=kwargs['all_tags'],
                                              schema=kwargs['schema'])
            for tag in tags:
                self.add_tag(tag)

            return

        # If we are reading from a CSV file, go ahead and parse it
        if 'csv' in kwargs and kwargs['csv']:
            csv_file = csv_reader(star_buffer)
            self.add_tag(next(csv_file))
            for row in csv_file:
                self.add_data(row, convert_data_types=kwargs.get('convert_data_types', False))
            self.source = f"from_csv('{kwargs['csv']}')"
            return

        tmp_entry = entry_mod.Entry.from_scratch(0)

        # Load the BMRB entry from the file
        star_buffer = StringIO(f"data_0 save_internaluseyoushouldntseethis_frame _internal.use internal "
                               f"{star_buffer.read()} save_")
        parser = Parser(entry_to_parse_into=tmp_entry)
        parser.parse(star_buffer.read(), source=self.source, convert_data_types=kwargs.get('convert_data_types', False),
                     raise_parse_warnings=kwargs.get('raise_parse_warnings', False))

        # Check that there was only one loop here
        if len(tmp_entry[0].loops) > 1:
            raise ValueError("You attempted to parse one loop but the source you provided had more than one loop. "
                             "Please either parse all loops as a saveframe or only parse one loop. Loops detected: " +
                             str(tmp_entry[0].loops))

        # Copy the first parsed saveframe into ourself
        self._tags = tmp_entry[0][0].tags
        self.data = tmp_entry[0][0].data
        self.category = tmp_entry[0][0].category

    def __iter__(self) -> list:
        """ Yields each of the rows contained within the loop. """

        for row in self.data:
            yield row

    def __len__(self) -> int:
        """Return the number of rows of data."""

        return len(self.data)

    def __lt__(self, other) -> bool:
        """Returns True if this loop sorts lower than the compared
        loop, false otherwise."""

        if not isinstance(other, Loop):
            return NotImplemented

        return self.category < other.category

    def __repr__(self) -> str:
        """Returns a description of the loop."""

        return f"<pynmrstar.Loop '{self.category}'>"

    def __setitem__(self, key: str, item: Any) -> None:
        """Set all of the instances of a tag to the provided value.
        If there are 5 rows of data in the loop, you will need to
        assign a list with 5 elements."""

        tag = utils.format_tag_lc(key)

        # Check that their tag is in the loop
        if tag not in self._lc_tags:
            raise ValueError(f"Cannot assign to tag '{key}' as it does not exist in this loop.")

        # Determine where to assign
        tag_id = self._lc_tags[tag]

        # Make sure they provide a list of the correct length
        if len(self[key]) != len(item):
            raise ValueError("To assign to a tag you must provide a list (or iterable) of a length equal to the "
                             f"number of values that currently exist for that tag. The tag '{key}' currently has"
                             f" {len(self[key])} values and you supplied {len(item)} values.")

        # Do the assignment
        for pos, row in enumerate(self.data):
            row[tag_id] = item[pos]

    def __str__(self, skip_empty_loops: bool = False, skip_empty_tags: bool = False) -> str:
        """Returns the loop in STAR format as a string."""

        # Check if there is any data in this loop
        if len(self.data) == 0:
            # They do not want us to print empty loops
            if skip_empty_loops:
                return ""
            else:
                # If we have no tags than return the empty loop
                if len(self._tags) == 0:
                    return "\n   loop_\n\n   stop_\n"

        if len(self._tags) == 0:
            raise InvalidStateError("Impossible to print data if there are no associated tags. Error in loop "
                                    f"'{self.category}' which contains data but hasn't had any tags added.")

        # Make sure the tags and data match
        self._check_tags_match_data()

        # If skipping null tags, it's easier to filter out a loop with only real tags and then print
        if skip_empty_tags:
            has_data = [not all([_ in definitions.NULL_VALUES for _ in column]) for column in zip(*self.data)]
            return self.filter([tag for x, tag in enumerate(self._tags) if has_data[x]]).format()

        # Start the loop
        return_chunks = ["\n   loop_\n"]
        # Print the tags
        format_string = "      %-s\n"

        # Check to make sure our category is set
        if self.category is None:
            raise InvalidStateError("The category was never set for this loop. Either add a tag with the category "
                                    "intact, specify it when generating the loop, or set it using Loop.set_category().")

        # Print the categories
        if self.category is None:
            for tag in self._tags:
                return_chunks.append(format_string % tag)
        else:
            for tag in self._tags:
                return_chunks.append(format_string % (self.category + "." + tag))

        return_chunks.append("\n")

        if len(self.data) != 0:

            # Make a copy of the data
            working_data = []
            title_widths = [4]*len(self.data[0])

            # Put quotes as needed on the data
            for row_pos, row in enumerate(self.data):
                clean_row = []
                for col_pos, x in enumerate(row):
                    try:
                        clean_val = utils.quote_value(x)
                        clean_row.append(clean_val)
                        length = len(clean_val) + 3
                        if length > title_widths[col_pos] and "\n" not in clean_val:
                            title_widths[col_pos] = length

                    except ValueError:
                        raise InvalidStateError('Cannot generate NMR-STAR for entry, as empty strings are not valid '
                                                'tag values in NMR-STAR. Please either replace the empty strings with'
                                                ' None objects, or set pynmrstar.definitions.STR_CONVERSION_DICT['
                                                '\'\'] = None.\n'
                                                f'Loop: {self.category} Row: {row_pos} Column: {col_pos}')

                working_data.append(clean_row)

            # Generate the format string
            format_string = "     " + "%-*s" * len(self._tags) + " \n"

            # Print the data, with the tags sized appropriately
            for datum in working_data:
                for pos, item in enumerate(datum):
                    if "\n" in item:
                        datum[pos] = "\n;\n%s;\n" % item

                # Print the data (combine the tags' widths with their data)
                tag_width_list = [d for d in zip(title_widths, datum)]
                return_chunks.append(format_string % tuple(chain.from_iterable(tag_width_list)))

        # Close the loop
        return "".join(return_chunks) + "\n   stop_\n"

    @property
    def _lc_tags(self) -> Dict[str, int]:
        return {_[1].lower(): _[0] for _ in enumerate(self._tags)}

    @property
    def empty(self) -> bool:
        """ Check if the loop has no data. """

        for row in self.data:
            for col in row:
                if col not in definitions.NULL_VALUES:
                    return False

        return True

    @property
    def tags(self) -> List[str]:
        return self._tags

    @classmethod
    def from_file(cls, the_file: Union[str, TextIO, BinaryIO], csv: bool = False, convert_data_types: bool = False,
                  raise_parse_warnings: bool = False):
        """Create a loop by loading in a file. Specify csv=True if
        the file is a CSV file. If the_file starts with http://,
        https://, or ftp:// then we will use those protocols to attempt
        to open the file.

        Setting convert_data_types to True will automatically convert
        the data loaded from the file into the corresponding python type as
        determined by loading the standard BMRB schema. This would mean that
        all floats will be represented as decimal.Decimal objects, all integers
        will be python int objects, strings and vars will remain strings, and
        dates will become datetime.date objects. When printing str() is called
        on all objects. Other that converting uppercase "E"s in scientific
        notation floats to lowercase "e"s this should not cause any change in
        the way re-printed NMR-STAR objects are displayed.

        Setting raise_parse_warnings to True will result in the raising of a
        ParsingError rather than logging a warning when non-valid (but
        ignorable) issues are found."""

        return cls(file_name=the_file, csv=csv, convert_data_types=convert_data_types,
                   raise_parse_warnings=raise_parse_warnings)

    @classmethod
    def from_json(cls, json_dict: Union[dict, str]):
        """Create a loop from JSON (serialized or unserialized JSON)."""

        # If they provided a string, try to load it using JSON
        if not isinstance(json_dict, dict):
            try:
                json_dict = json.loads(json_dict)
            except (TypeError, ValueError):
                raise ValueError("The JSON you provided was neither a Python dictionary nor a JSON string.")

        # Make sure it has the correct keys
        for check in ['tags', 'category', 'data']:
            if check not in json_dict:
                raise ValueError(f"The JSON you provide must be a dictionary and must contain the key '{check}' - even"
                                 f" if the key points to None.")

        # Create a loop from scratch and populate it
        ret = Loop.from_scratch()
        ret._tags = json_dict['tags']
        ret.category = json_dict['category']
        ret.data = json_dict['data']
        ret.source = "from_json()"

        # Return the new loop
        return ret

    @classmethod
    def from_scratch(cls, category: str = None, source: str = "from_scratch()"):
        """Create an empty saveframe that you can programmatically add
        to. You may also pass the tag prefix as the second argument. If
        you do not pass the tag prefix it will be set the first time you
        add a tag."""

        return cls(category=category, source=source)

    @classmethod
    def from_string(cls, the_string: str, csv: bool = False, convert_data_types: bool = False,
                    raise_parse_warnings: bool = False):
        """Create a loop by parsing a string. Specify csv=True is
        the string is in CSV format and not NMR-STAR format.

        Setting convert_data_types to True will automatically convert
        the data loaded from the file into the corresponding python type as
        determined by loading the standard BMRB schema. This would mean that
        all floats will be represented as decimal.Decimal objects, all integers
        will be python int objects, strings and vars will remain strings, and
        dates will become datetime.date objects. When printing str() is called
        on all objects. Other that converting uppercase "E"s in scientific
        notation floats to lowercase "e"s this should not cause any change in
        the way re-printed NMR-STAR objects are displayed.

        Setting raise_parse_warnings to True will result in the raising of a
        ParsingError rather than logging a warning when non-valid (but
        ignorable) issues are found."""

        return cls(the_string=the_string, csv=csv, convert_data_types=convert_data_types,
                   raise_parse_warnings=raise_parse_warnings)

    @classmethod
    def from_template(cls, tag_prefix: str, all_tags: bool = False, schema: Schema = None):
        """ Create a loop that has all of the tags from the schema present.
        No values will be assigned. Specify the tag prefix of the loop.

        The optional argument all_tags forces all tags to be included
        rather than just the mandatory tags."""

        schema = utils.get_schema(schema)
        return cls(tag_prefix=tag_prefix, all_tags=all_tags,
                   schema=schema, source=f"from_template({schema.version})")

    @staticmethod
    def _get_tags_from_schema(category: str, schema: Schema = None, all_tags: bool = False) -> List[str]:
        """ Returns the tags from the schema for the category of this
        loop. """

        schema = utils.get_schema(schema)

        # Put the _ on the front for them if necessary
        if not category.startswith("_"):
            category = "_" + category
        if not category.endswith("."):
            category = category + "."

        tags = []

        for item in schema.schema_order:
            # The tag is in the loop
            if item.lower().startswith(category.lower()):

                # Unconditional add
                if all_tags:
                    tags.append(item)
                # Conditional add
                else:
                    if schema.schema[item.lower()]["public"] != "I":
                        tags.append(item)
        if len(tags) == 0:
            raise InvalidStateError(f"The tag prefix '{category}' has no corresponding tags in the dictionary.")

        return tags

    def _check_tags_match_data(self) -> bool:
        """ Ensures that each row of the data has the same number of
        elements as there are tags for the loop. This is necessary to
        print or do some other operations on loops that count on the values
        matching. """

        # Make sure that if there is data, it is the same width as the
        #  tag names
        if len(self.data) > 0:
            for x, row in enumerate(self.data):
                if len(self._tags) != len(row):
                    raise InvalidStateError(f"The number of tags must match the width of the data. Error in loop "
                                            f"'{self.category}'. In this case, there are {len(self._tags)} tags, and "
                                            f"row number {x} has {len(row)} tags.")

        return True

    def add_data(self,
                 data: Union[List[dict], Dict[str, List], List[Union[str, float, int]], List[List[Any]]],
                 rearrange: bool = False,
                 convert_data_types: bool = False):
        """Add data to a loop. You can provide the data to add organized in four different ways, though the first
        two are recommended for new code. The other two (#3 and #4) are preserved for sake of existing code (written
        prior to version 3.3) and for niche use cases:

        1: You can provide a list of dictionaries of tags to add. For example,
        ``[{'name': 'Jeff', 'location': 'Connecticut'}, {'name': 'Chad', 'location': 'Madison'}]`` will add two new
        rows, and set the values of the tags ``name`` and ``location`` to the values provided. If there are other
        tags in the loop, they will be assigned null values for the rows corresponding to the tags added.

        2: You can provide a dictionary of lists, as such (corresponds to adding the same ultimate data as in the
        example #1): ``{'name': ['Jeff', 'Chad'], 'location': ['Connecticut', 'Madison']}``. This will also create
        two new rows in the loop and assign the values provided.
        
        3: You can provide a list of lists of tag values to add. In this case, each list must have the same tag
        values (and order of tags) as the known tags present in the loop. To correspond to the above examples, the data
        would look like: ``[['Jeff', 'Connecticut'], ['Chad', 'Madison']]``. Adding data this way requires both that
        you provide values for all tags present in the loop, and that you provide the values in the same order that the
        tags already are already defined in the loop.

        4. You can provide a single list of tag values to add. In the most simple case, that would correspond to just
        adding one row of data in the same was as in #3 above, as such: ``['Jeff', 'Connecticut']``. In a more
        complicated example, you could also add data (corresponding to example #1 and #2) as such:
        ``['Jeff', 'Connecticut', 'Chad', 'Madison']`` - but if you provide data this way, you must set
         ``rearrange=True``. This usage is strongly discouraged, but exists for legacy reasons.

        :param data: See the docstring for the method.
        :type data: Union[List[dict], Dict[str, List], List[Union[str, float, int]], List[List[Any]]]
        :param convert_data_types: If true, converts data you provide into the data type defined in the dictionary.
            For example, if you provided the string '5' for the tag ``_Atom_chem_shift.Val``, it would automatically
            be converted to a float while being added. This is mainly useful for parsers, as your data is probably
            already in a format that is usable for you.
        :type convert_data_types: bool
        :param rearrange: If true, rearrange data provided in method #4 as necessary to fit in the loop. This only
            exists for parsers, and it's use is strongly discouraged.
        :type rearrange: bool
        """

        if not data:
            raise ValueError('No valid data provided.')

        pending_data: List = []
        lc_tag_index: Dict[str, int] = self._lc_tags

        def format_two_to_one(format_two: Dict[str, List]):
            max_length = max([len(_) for _ in format_two.values()])
            keys = format_two.keys()
            for row_id in range(0, max_length):
                row_dict = {}
                for key in keys:
                    try:
                        row_dict[key] = format_two[key][row_id]
                    except IndexError:
                        pass
                yield row_dict

        # Data format #1 and #2
        if (isinstance(data, list) and isinstance(data[0], dict)) or \
                isinstance(data, dict) and all([isinstance(_, list) for _ in data.values()]):

            # Handle format #2 by converting it to #1
            if isinstance(data, dict):
                data = format_two_to_one(data)

            for pos, row in enumerate(data):
                current_row = [None]*len(self._tags)
                for tag, value in row.items():
                    try:
                        tag_index = lc_tag_index[utils.format_tag_lc(tag)]
                    except KeyError:
                        raise ValueError(f'In row {pos} of your provided data, a tag was supplied which was not'
                                         f" already present in the loop. Invalid tag: '{tag}'")
                    current_row[tag_index] = value
                pending_data.append(current_row)
        # Type 4 - a list of lists
        elif isinstance(data, list) and isinstance(data[0], list):
            for pos, row in enumerate(data):
                if len(row) != len(self.tags):
                    raise ValueError('One of the lists you provided is not the correct length to match the number '
                                     f'of tags present in the loop. Error on row {pos} with values: {row}')
            pending_data = data
        # Type 3 - a list of values
        elif isinstance(data, list):
            if rearrange:
                # Break their data into chunks based on the number of tags
                pending_data = [data[x:x + len(self._tags)] for x in range(0, len(data), len(self._tags))]
                if len(pending_data[-1]) != len(self._tags):
                    raise ValueError(f"The number of data elements in the list you provided is not an even multiple of "
                                     f"the number of tags which are set in the loop. Please either add missing tags "
                                     f"using Loop.add_tag() or modify the list of tag values you are adding to be an "
                                     f"even multiple of the number of tags. Error in loop '{self.category}'.")
            else:
                # Add one row of data
                if len(data) != len(self._tags):
                    raise ValueError("The list must have the same number of elements as the number of tags when adding "
                                     "a single row of values! Insert tag names first by calling Loop.add_tag().")
                # Add the user data
                pending_data.append(data)
        else:
            raise ValueError("Your data did not match one of the supported types.")

        # Auto convert data types if option set
        if convert_data_types:
            schema = utils.get_schema()
            for row in pending_data:
                for tag_id, datum in enumerate(row):
                    row[tag_id] = schema.convert_tag(f"{self.category}.{self._tags[tag_id]}", datum)

        # Add the data at the very end to ensure that errors are caught before we mutate the data
        self.data.extend(pending_data)

    def add_data_by_tag(self, tag_name: str, value) -> None:
        """Deprecated: It is recommended to use add_data() instead for most use
        cases.

        Add data to the loop one element at a time, based on tag.
        Useful when adding data from SANS parsers."""

        warnings.warn("Deprecated: It is recommended to use Loop.add_data() instead for most use cases.",
                      DeprecationWarning)

        # Make sure the category matches - if provided
        if "." in tag_name:
            supplied_category = utils.format_category(str(tag_name))
            if supplied_category.lower() != self.category.lower():
                raise ValueError(f"Category provided in your tag '{supplied_category}' does not match this loop's "
                                 f"category '{self.category}'.")

        pos = self.tag_index(tag_name)
        if pos is None:
            raise ValueError(f"The tag '{tag_name}' to which you are attempting to add data does not yet exist. Create "
                             f"the tags using Loop.add_tag() before adding data.")
        if len(self.data) == 0:
            self.data.append([])
        if len(self.data[-1]) == len(self._tags):
            self.data.append([])
        if len(self.data[-1]) != pos:
            raise ValueError("You cannot add data out of tag order.")
        self.data[-1].append(value)

    def add_missing_tags(self, schema: 'Schema' = None, all_tags: bool = False) -> None:
        """ Automatically adds any missing tags (according to the schema),
        sorts the tags, and renumbers the tags by ordinal. """

        self.add_tag(Loop._get_tags_from_schema(self.category, schema=schema, all_tags=all_tags),
                     ignore_duplicates=True, update_data=True)
        self.sort_tags()

        # See if we can sort the rows (in addition to tags)
        try:
            self.sort_rows("Ordinal")
        except ValueError:
            pass
        except TypeError:
            ordinal_idx = self.tag_index("Ordinal")

            # If we are in another row, assign to the previous row
            for pos, row in enumerate(self.data):
                row[ordinal_idx] = pos + 1

    def add_tag(self, name: Union[str, List[str]], ignore_duplicates: bool = False, update_data: bool = False) -> None:
        """Add a tag to the tag name list. Does a bit of validation
        and parsing. Set ignore_duplicates to true to ignore attempts
        to add the same tag more than once rather than raise an
        exception.

        You can also pass a list of tag names to add more than one
        tag at a time.

        Adding a tag will update the data array to match by adding
        None values to the rows if you specify update_data=True."""

        # If they have passed multiple tags to add, call ourself
        #  on each of them in succession
        if isinstance(name, (list, tuple)):
            for item in name:
                self.add_tag(item, ignore_duplicates=ignore_duplicates, update_data=update_data)
            return

        name = name.strip()

        if "." in name:
            if name[0] != ".":
                category = name[0:name.index(".")]
                if category[:1] != "_":
                    category = "_" + category

                if self.category is None:
                    self.category = category
                elif self.category.lower() != category.lower():
                    raise ValueError("One loop cannot have tags with different categories (or tags that don't "
                                     f"match the loop category)! The loop category is '{self.category}' while "
                                     f"the category in the tag was '{category}'.")
                name = name[name.index(".") + 1:]
            else:
                name = name[1:]

        # Ignore duplicate tags
        if self.tag_index(name) is not None:
            if ignore_duplicates:
                return
            else:
                raise ValueError(f"There is already a tag with the name '{name}' in the loop '{self.category}'.")
        if name in definitions.NULL_VALUES:
            raise ValueError(f"Cannot use a null-equivalent value as a tag name. Invalid tag name: '{name}'")
        if "." in name:
            raise ValueError(f"There cannot be more than one '.' in a tag name. Invalid tag name: '{name}'")
        for char in str(name):
            if char in utils.definitions.WHITESPACE:
                raise ValueError(f"Tag names can not contain whitespace characters. Invalid tag name: '{name}")

        # Add the tag
        self._tags.append(name)

        # Add None's to the rows of data
        if update_data:

            for row in self.data:
                row.append(None)

    def clear_data(self) -> None:
        """Erases all data in this loop. Does not erase the tag names
        or loop category."""

        self.data = []

    def compare(self, other) -> List[str]:
        """Returns the differences between two loops as a list. Order of
        loops being compared does not make a difference on the specific
        errors detected."""

        diffs = []

        # Check if this is literally the same object
        if self is other:
            return []
        # Check if the other object is our string representation
        if isinstance(other, str):
            if str(self) == other:
                return []
            else:
                return ['String was not exactly equal to loop.']
        elif not isinstance(other, Loop):
            return ['Other object is not of class Loop.']

        # We need to do this in case of an extra "\n" on the end of one tag
        if str(other) == str(self):
            return []

        # Do STAR comparison
        try:
            # Check category of loops
            if str(self.category).lower() != str(other.category).lower():
                diffs.append(f"\t\tCategory of loops does not match: '{self.category}' vs '{other.category}'.")

            # Check tags of loops
            if ([x.lower() for x in self._tags] !=
                    [x.lower() for x in other.tags]):
                diffs.append(f"\t\tLoop tag names do not match for loop with category '{self.category}'.")

            # No point checking if data is the same if the tag names aren't
            else:
                # Only sort the data if it is not already equal
                if self.data != other.data:

                    # Check data of loops
                    self_data = sorted(deepcopy(self.data))
                    other_data = sorted(deepcopy(other.data))

                    if self_data != other_data:
                        diffs.append(f"\t\tLoop data does not match for loop with category '{self.category}'.")

        except AttributeError as err:
            diffs.append(f"\t\tAn exception occurred while comparing: '{err}'.")

        return diffs

    def delete_tag(self, tag: Union[str, List[str]]) -> None:
        """ Deprecated. Please use `py:meth:pynmrstar.Loop.remove_tag` instead. """

        warnings.warn('Please use remove_tag() instead.', DeprecationWarning)
        return self.remove_tag(tag)

    def delete_data_by_tag_value(self, tag: str, value: Any, index_tag: str = None) -> List[List[Any]]:
        """ Deprecated. Please use `py:meth:pynmrstar.Loop.remove_data_by_tag_value` instead. """

        warnings.warn('Please use remove_data_by_tag_value() instead.', DeprecationWarning)
        return self.remove_data_by_tag_value(tag, value, index_tag)

    def filter(self, tag_list: Union[str, List[str], Tuple[str]], ignore_missing_tags: bool = False):
        """ Returns a new loop containing only the specified tags.
        Specify ignore_missing_tags=True to bypass missing tags rather
        than raising an error."""

        result = Loop.from_scratch()
        valid_tags = []

        # If they only provide one tag make it a list
        if not isinstance(tag_list, (list, tuple)):
            tag_list = [tag_list]

        # Make sure all the tags specified exist
        for tag in tag_list:

            # Handle an invalid tag
            tag_match_index = self.tag_index(tag)
            if tag_match_index is None:
                if not ignore_missing_tags:
                    raise KeyError(f"Cannot filter tag '{tag}' as it isn't present in this loop.")
                continue

            valid_tags.append(tag)
            result.add_tag(self._tags[tag_match_index])

        # Add the data for the tags to the new loop
        results = self.get_tag(valid_tags)

        # If there is only a single tag, we can't add data the same way
        if len(valid_tags) == 1:
            for item in results:
                result.add_data([item])
        else:
            for row in results:
                # We know it's a row because we didn't specify dict_result=True to get_tag()
                assert isinstance(row, list)
                result.add_data(row)

        # Assign the category of the new loop
        if result.category is None:
            result.category = self.category

        return result

    def format(self, skip_empty_loops: bool = True, skip_empty_tags: bool = False) -> str:
        """ The same as calling str(Loop), except that you can pass options
        to customize how the loop is printed.

        skip_empty_loops will omit printing loops with no tags at all. (A loop with null tags is not "empty".)
        skip_empty_tags will omit tags in the loop which have no non-null values."""

        return self.__str__(skip_empty_loops=skip_empty_loops, skip_empty_tags=skip_empty_tags)

    def get_data_as_csv(self, header: bool = True, show_category: bool = True) -> str:
        """Return the data contained in the loops, properly CSVd, as a
        string. Set header to False to omit the header. Set
        show_category to false to omit the loop category from the
        headers."""

        csv_buffer = StringIO()
        csv_writer_object = csv_writer(csv_buffer)

        if header:
            if show_category:
                csv_writer_object.writerow(
                    [str(self.category) + "." + str(x) for x in self._tags])
            else:
                csv_writer_object.writerow([str(x) for x in self._tags])

        for row in self.data:

            data = []
            for piece in row:
                data.append(piece)

            csv_writer_object.writerow(data)

        csv_buffer.seek(0)
        return csv_buffer.read().replace('\r\n', '\n')

    def get_json(self, serialize: bool = True) -> Union[dict, str]:
        """ Returns the loop in JSON format. If serialize is set to
        False a dictionary representation of the loop that is
        serializeable is returned."""

        loop_dict = {
            "category": self.category,
            "tags": self._tags,
            "data": self.data
        }

        if serialize:
            return json.dumps(loop_dict, default=_json_serialize)
        else:
            return loop_dict

    def get_tag_names(self) -> List[str]:
        """ Return the tag names for this entry with the category
        included. Throws ValueError if the category was never set.

        To fetch tag values use get_tag()."""

        if not self.category:
            raise InvalidStateError("You never set the category of this loop. You must set the category before calling "
                                    "this method, either by setting the loop category directly when creating the loop "
                                    "using the Loop.from_scratch() class method, by calling loop.set_category(), or by "
                                    "adding a fully qualified tag which includes the loop category (for example, "
                                    "adding '_Citation_author.Family_name' rather than just 'Family_name').")

        return [self.category + "." + x for x in self._tags]

    def get_tag(self, tags: Optional[Union[str, List[str]]] = None, whole_tag: bool = False,
                dict_result: bool = False) -> Union[List[Any], List[Dict[str, Any]]]:
        """Provided a tag name (or a list of tag names) return the selected tags by row as
        a list of lists.

        If whole_tag=True return the full tag name along with the tag
        value, or if dict_result=True, as the tag key.

        If dict_result=True, return the tags as a list of dictionaries
        in which the tag value points to the tag."""

        # All tags
        if tags is None:
            if not dict_result:
                return self.data
            else:
                tags = [self._tags]
        # Turn single elements into lists
        if not isinstance(tags, list):
            tags = [tags]

        # Make a copy of the tags to fetch - don't modify the
        # list that was passed
        lower_tags = deepcopy(tags)

        # Strip the category if they provide it (also validate
        #  it during the process)
        for pos, item in enumerate([str(x) for x in lower_tags]):
            if "." in item and utils.format_category(item).lower() != self.category.lower():
                raise ValueError(f"Cannot fetch data with tag '{item}' because the category does not match the "
                                 f"category of this loop '{self.category}'.")
            lower_tags[pos] = utils.format_tag_lc(item)

        # Make a lower case copy of the tags
        tags_lower = [x.lower() for x in self._tags]

        # Map tag name to tag position in list
        tag_mapping = dict(zip(reversed(tags_lower), reversed(range(len(tags_lower)))))

        # Make sure their fields are actually present in the entry
        tag_ids = []
        for pos, query in enumerate(lower_tags):
            if str(query) in tag_mapping:
                tag_ids.append(tag_mapping[query])
            elif isinstance(query, int):
                tag_ids.append(query)
            else:
                raise KeyError(f"Could not locate the tag with name or ID: '{tags[pos]}' in loop '{self.category}'.")

        # First build the tags as a list
        if not dict_result:

            # Use a list comprehension to pull the correct tags out of the rows
            if whole_tag:
                result = [[[self.category + "." + self._tags[col_id], row[col_id]]
                           for col_id in tag_ids] for row in self.data]
            else:
                result = [[row[col_id] for col_id in tag_ids] for row in self.data]

            # Strip the extra list if only one tag
            if len(lower_tags) == 1:
                return [x[0] for x in result]
            else:
                return result
        # Make a dictionary
        else:
            if whole_tag:
                result = [dict((self.category + "." + self._tags[col_id], row[col_id]) for col_id in tag_ids) for
                          row in self.data]
            else:
                result = [dict((self._tags[col_id], row[col_id]) for col_id in tag_ids) for row in self.data]

        return result

    def print_tree(self) -> None:
        """Prints a summary, tree style, of the loop."""

        print(repr(self))

    def remove_data_by_tag_value(self, tag: str, value: Any, index_tag: str = None) -> List[List[Any]]:
        """Removes all rows which contain the provided value in the
        provided tag name. If index_tag is provided, that tag is
        renumbered starting with 1. Returns the deleted rows."""

        # Make sure the category matches - if provided
        if "." in tag:
            supplied_category = utils.format_category(str(tag))
            if supplied_category.lower() != self.category.lower():
                raise ValueError(f"The category provided in your tag '{supplied_category}' does not match this loop's "
                                 f"category '{self.category}'.")

        search_tag = self.tag_index(tag)
        if search_tag is None:
            raise ValueError(f"The tag you provided '{tag}' isn't in this loop!")

        deleted = []

        # Delete all rows in which the user-provided tag matched
        cur_row = 0
        while cur_row < len(self.data):
            if self.data[cur_row][search_tag] == value:
                deleted.append(self.data.pop(cur_row))
                continue
            cur_row += 1

        # Re-number if they so desire
        if index_tag is not None:
            self.renumber_rows(index_tag)

        return deleted

    def remove_tag(self, tag: Union[str, List[str]]) -> None:
        """Removes one or more tags from the loop based on tag name. Also removes any data for the given tag.
        Provide either a tag or list of tags."""

        if not isinstance(tag, list):
            tag = [tag]

        # Check if the tags exist first
        for each_tag in tag:
            if self.tag_index(each_tag) is None:
                raise KeyError(f"There is no tag with name '{each_tag}' to remove in loop '{self.category}'.")

        # Calculate the tag position each time, because it will change as the previous tag is deleted
        for each_tag in tag:
            tag_position: int = self.tag_index(each_tag)
            del self._tags[tag_position]
            for row in self.data:
                del row[tag_position]

    def renumber_rows(self, index_tag: str, start_value: int = 1, maintain_ordering: bool = False):
        """Renumber a given tag incrementally. Set start_value to
        initial value if 1 is not acceptable. Set maintain_ordering to
        preserve sequence with offset.

        E.g. 2,3,3,5 would become 1,2,2,4."""

        # Make sure the category matches
        if "." in str(index_tag):
            supplied_category = utils.format_category(str(index_tag))
            if supplied_category.lower() != self.category.lower():
                raise ValueError(f"Category provided in your tag '{supplied_category}' does not match this loop's "
                                 f"category '{self.category}'.")

        # Determine which tag ID to renumber
        renumber_tag = self.tag_index(index_tag)

        # The tag to replace in is the tag they specify
        if renumber_tag is None:
            # Or, perhaps they specified an integer to represent the tag?
            try:
                renumber_tag = int(index_tag)
            except ValueError:
                raise ValueError(f"The renumbering tag you provided '{index_tag}' isn't in this loop!")

        # Do nothing if we have no data
        if len(self.data) == 0:
            return

        # Make sure the tags and data match
        self._check_tags_match_data()

        if maintain_ordering:
            # If they have a string buried somewhere in the row, we'll
            #  have to restore the original values
            data_copy = deepcopy(self.data)
            offset = 0
            for pos in range(0, len(self.data)):
                try:
                    if pos == 0:
                        offset = start_value - int(self.data[0][renumber_tag])
                    new_data = int(self.data[pos][renumber_tag]) + offset

                    if isinstance(self.data[pos][renumber_tag], str):
                        self.data[pos][renumber_tag] = str(new_data)
                    else:
                        self.data[pos][renumber_tag] = new_data
                except ValueError:
                    self.data = data_copy
                    raise ValueError("You can't renumber a row containing anything that can't be coerced into an "
                                     "integer using maintain_ordering. I.e. what am I suppose to renumber "
                                     f"'{self.data[pos][renumber_tag]}' to?")

        # Simple renumbering algorithm if we don't need to maintain the ordering
        else:
            for pos in range(0, len(self.data)):
                if isinstance(self.data[pos][renumber_tag], str):
                    self.data[pos][renumber_tag] = str(pos + start_value)
                else:
                    self.data[pos][renumber_tag] = pos + start_value

    def set_category(self, category: str) -> None:
        """ Set the category of the loop. Useful if you didn't know the
        category at loop creation time."""

        self.category = utils.format_category(category)

    def sort_tags(self, schema: 'Schema' = None) -> None:
        """ Rearranges the tag names and data in the loop to match the order
        from the schema. Uses the BMRB schema unless one is provided."""

        schema = utils.get_schema(schema)
        current_order = self.get_tag_names()

        # Sort the tags
        def sort_key(_) -> int:
            return schema.tag_key(_)

        sorted_order = sorted(current_order, key=sort_key)

        # Don't touch the data if the tags are already in order
        if sorted_order == current_order:
            return
        else:
            self.data = self.get_tag(sorted_order)
            self._tags = [utils.format_tag(x) for x in sorted_order]

    def sort_rows(self, tags: Union[str, List[str]], key: Callable = None) -> None:
        """ Sort the data in the rows by their values for a given tag
        or tags. Specify the tags using their names or ordinals.
        Accepts a list or an int/float. By default we will sort
        numerically. If that fails we do a string sort. Supply a
        function as key and we will order the elements based on the
        keys it provides. See the help for sorted() for more details. If
        you provide multiple tags to sort by, they are interpreted as
        increasing order of sort priority."""

        # Do nothing if we have no data
        if len(self.data) == 0:
            return

        # This will determine how we sort
        sort_ordinals = []

        if isinstance(tags, list):
            processing_list = tags
        else:
            processing_list = [tags]

        # Process their input to determine which tags to operate on
        for cur_tag in [str(x) for x in processing_list]:

            # Make sure the category matches
            if "." in cur_tag:
                supplied_category = utils.format_category(cur_tag)
                if supplied_category.lower() != self.category.lower():
                    raise ValueError(f"The category provided in your tag '{supplied_category}' does not match this "
                                     f"loop's category '{self.category}'.")

            renumber_tag = self.tag_index(cur_tag)

            # They didn't specify a valid tag
            if renumber_tag is None:
                # Perhaps they specified an integer to represent the tag?
                try:
                    renumber_tag = int(cur_tag)
                except ValueError:
                    raise ValueError(f"The sorting tag you provided '{cur_tag}' isn't in this loop!")

            sort_ordinals.append(renumber_tag)

        # Do the sort(s)
        for tag in sort_ordinals:
            # Going through each tag, first attempt to sort as integer.
            #  Then fallback to string sort.
            try:
                if key is None:
                    tmp_data = sorted(self.data, key=lambda _, pos=tag: float(_[pos]))
                else:
                    tmp_data = sorted(self.data, key=key)
            except ValueError:
                if key is None:
                    tmp_data = sorted(self.data, key=lambda _, pos=tag: _[pos])
                else:
                    tmp_data = sorted(self.data, key=key)
            self.data = tmp_data

    def tag_index(self, tag_name: str) -> Optional[int]:
        """ Helper method to do a case-insensitive check for the presence
        of a given tag in this loop. Returns the index of the tag if found
        and None if not found.

        This is useful if you need to get the index of a certain tag to
        iterate through the data and modify it."""

        try:
            return self._lc_tags[utils.format_tag_lc(str(tag_name))]
        except KeyError:
            return None

    def validate(self, validate_schema: bool = True, schema: 'Schema' = None,
                 validate_star: bool = True, category: str = None) -> List[str]:
        """Validate a loop in a variety of ways. Returns a list of
        errors found. 0-length list indicates no errors found. By
        default all validation modes are enabled.

        validate_schema - Determines if the entry is validated against
        the NMR-STAR schema. You can pass your own custom schema if desired,
        otherwise the schema will be fetched from the BMRB servers.

        validate_star - Determines if the STAR syntax checks are ran."""

        errors = []

        if validate_schema:
            # Get the default schema if we are not passed a schema
            my_schema = utils.get_schema(schema)

            # Check the data
            for row_num, row in enumerate(self.data):
                for pos, datum in enumerate(row):
                    errors.extend(my_schema.val_type(f"{self.category}.{self._tags[pos]}", datum, category=category))

        if validate_star:
            # Check for wrong data size
            num_cols = len(self._tags)
            for row_num, row in enumerate(self.data):
                # Make sure the width matches
                if len(row) != num_cols:
                    errors.append(f"Loop '{self.category}' data width does not match it's tag width on "
                                  f"row '{row_num}'.")

        return errors



#!/usr/bin/env python3

import os
from setuptools import setup, Extension


def get_version():
    internal_file_location = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'pynmrstar', '_internal.py')

    with open(internal_file_location, 'r') as internal_file:
        for line in internal_file:
            if line.startswith('__version__'):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
        else:
            raise RuntimeError("Unable to find version string.")


# Should fail if the readme is missing
long_des = open('README.rst', 'r').read()

cnmrstar = Extension('cnmrstar',
                     sources=['c/cnmrstarmodule.c'],
                     extra_compile_args=["-funroll-loops", "-O3"],
                     optional=True)

setup(name='pynmrstar',
      version=get_version(),
      packages=['pynmrstar'],
      ext_modules=[cnmrstar],
      install_requires=['requests>=2.21.0,<=3'],
      python_requires='>=3.6',
      author='Jon Wedell',
      author_email='wedell@uchc.edu',
      description='PyNMR-STAR provides tools for reading, writing, modifying, and interacting with NMR-STAR files. '
                  'Maintained by the BMRB.',
      long_description=long_des,
      long_description_content_type='text/x-rst',
      keywords=['bmrb', 'parser', 'nmr', 'nmrstar', 'biomagresbank', 'biological magnetic resonance bank'],
      url='https://github.com/uwbmrb/PyNMRSTAR',
      license='MIT',
      package_data={'pynmrstar': ['reference_files/schema.csv',
                                  'reference_files/comments.str',
                                  'reference_files/data_types.csv']},
      classifiers=[
          'Development Status :: 6 - Mature',
          'Environment :: Console',
          'Programming Language :: Python :: 3 :: Only',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Natural Language :: English',
          'Operating System :: MacOS',
          'Operating System :: POSIX :: Linux',
          'Operating System :: Microsoft :: Windows',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules'
      ]
      )
